"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = __importStar(require("@oclif/command"));
const promise_1 = __importDefault(require("simple-git/promise"));
const utils_1 = require("../lib/utils");
const common_tags_1 = require("common-tags");
const Git = __importStar(require("../lib/git"));
const Output = __importStar(require("../lib/output"));
const Publish = __importStar(require("../lib/publish"));
const Semver = __importStar(require("../lib/semver"));
const Context = __importStar(require("../lib/context"));
class Preview extends command_1.default {
    run() {
        return __awaiter(this, void 0, void 0, function* () {
            const { flags } = this.parse(Preview);
            const send = createOutputters({ json: flags.json });
            const git = promise_1.default();
            // TODO validate for missing or faulty package.json
            // TODO validate for dirty git status
            // TODO validate for found releases that fall outside the subset we support.
            // For example #.#.#-foobar.1 is something we would not know what to do
            // with. A good default is probably to hard-error when these are
            // encountered. But offer a flag/config option called e.g. "ignore_unsupport_pre_release_identifiers"
            // TODO handle edge case: not a git repo
            // TODO handle edge case: a git repo with no commits
            // TODO nicer tag rendering:
            //    1. for annotated tags should the messge
            //    2. show the tag author name
            //    3. show the the date the tag was made
            const ctx = yield Context.scan({
                overrides: {
                    trunk: flags.trunk || null,
                },
            });
            if (ctx.series.current.releases.stable ||
                ctx.series.current.releases.preview) {
                return send.commitAlreadyPreAndOrStableReleased(ctx.series.current);
            }
            if (ctx.currentBranch.isTrunk) {
                if (flags['show-type']) {
                    return send.releaseType({ type: 'stable', reason: 'is_trunk' });
                }
                const nextRelease = yield calcNextStablePreview(ctx.series);
                if (nextRelease === null) {
                    return send.noReleaseToMake();
                }
                if (flags['dry-run']) {
                    return send.dryRun(nextRelease);
                }
                yield Publish.publish({
                    distTag: 'next',
                    version: nextRelease.nextVersion,
                });
                // force update so the tag moves to a new commit
                yield git.raw(['tag', '-f', 'next']);
                yield git.raw(['push', 'origin', ':refs/tags/next']);
                yield git.raw(['push', '--tags']);
                console.log('updated git-tag "next"');
                return;
            }
            // todo refator with ctx data
            const prCheck = yield Git.checkBranchPR(git);
            if (prCheck.isPR === false) {
                return send.invalidBranchForPreRelease();
            }
            if (flags['show-type']) {
                return send.releaseType({
                    type: 'pr',
                    reason: prCheck.inferredBy,
                });
            }
            // TODO
            return process.stdout.write('todo: pr preview release');
        });
    }
}
exports.Preview = Preview;
Preview.flags = {
    trunk: command_1.flags.string({
        default: '',
        description: 'State which branch is trunk. Defaults to honuring the "base" branch setting in the GitHub repo settings.',
    }),
    /**
     * This flag is mostly used for debugging. It allows the user to see what
     * kind of preview release _would_ be made under the current conditions, and
     * why.
     */
    'show-type': command_1.flags.boolean({
        default: false,
        description: 'output the kind of preview release that would be made and why',
    }),
    'dry-run': command_1.flags.boolean({
        default: false,
        description: 'output what the next version would be if released now',
    }),
    json: command_1.flags.boolean({
        default: false,
        description: 'format output as JSON',
    }),
};
/**
 * 1. Find the last pre-release on the current branch. Take its build number. If
 *    none use 1.
 * 2. Calculate the semver bump type. Do this by analyizing the commits on the
 *    branch between HEAD and the last stable git tag.  The highest change type
 *    found is used. If no previous stable git tag use 0.0.1.
 * 3. Bump last stable version by bump type, thus producing the next version.
 * 4. Construct new version {nextVer}-next.{buildNum}. Example: 1.2.3-next.1.
 */
function calcNextStablePreview(series) {
    // We need all the commits since the last stable release to calculate the
    // pre-release. The pre-release is a bump against the last stable plus a
    // build number. The bump type used to bump is based on the aggregate of
    // all commits since the last stable. While the build number always
    // increments the maj/min/pat may only increment upon the first
    // pre-release, unless a later pre-release incurs a higher-order bump-type
    // e.g. begin with patch-kind changes followed later by min-kind changes.
    var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k, _l, _m;
    // Calculate the next version
    const incType = Semver.calcIncType(series.commitsSinceStable.map(c => c.message));
    if (incType === null)
        return null;
    const nextStable = Semver.incStable(incType, (_b = (_a = series.previousStable) === null || _a === void 0 ? void 0 : _a.releases.stable, (_b !== null && _b !== void 0 ? _b : Semver.zeroVer)));
    const nextVer = Semver.stableToPreview(nextStable, 'next', (_d = (_c = series.previousPreview) === null || _c === void 0 ? void 0 : _c.releases.preview.preRelease.buildNum, (_d !== null && _d !== void 0 ? _d : Semver.zeroBuildNum)) + 1);
    // TODO simplify by exposing series struct as is, little value in the remix here
    return {
        currentStable: (_f = (_e = series.previousStable) === null || _e === void 0 ? void 0 : _e.releases.stable.version, (_f !== null && _f !== void 0 ? _f : null)),
        currentPreviewNumber: (_h = (_g = series.previousPreview) === null || _g === void 0 ? void 0 : _g.releases.preview.preRelease.buildNum, (_h !== null && _h !== void 0 ? _h : null)),
        nextStable: nextStable.version,
        nextPreviewNumber: nextVer.preRelease.buildNum,
        currentVersion: (_m = (_k = (_j = series.previousPreview) === null || _j === void 0 ? void 0 : _j.releases.preview.version, (_k !== null && _k !== void 0 ? _k : (_l = series.previousStable) === null || _l === void 0 ? void 0 : _l.releases.stable.version)), (_m !== null && _m !== void 0 ? _m : null)),
        nextVersion: nextVer.version,
        commitsInRelease: series.previousPreview
            ? series.commitsSincePreview.map(c => c.message)
            : series.commitsSinceStable.map(c => c.message),
        bumpType: incType,
        isFirstVer: series.previousStable === null && series.previousPreview === null,
        isFirstVerStable: series.previousStable === null,
        isFirstVerPreRelease: series.previousPreview === null,
    };
}
function createOutputters(opts) {
    /**
     * Output the release type.
     */
    function releaseType(info) {
        if (opts.json) {
            Output.outputOk('release_type', info);
        }
        else {
            Output.outputRaw(`The release type that would be made right now is ${info.type} becuase ${info.reason}.`);
        }
    }
    /**
     * Output no release to make notice.
     */
    function noReleaseToMake() {
        Output.output(Output.createException('no_release_to_make', {
            summary: 'All commits are either meta or not conforming to conventional commit. No release will be made.',
        }), { json: opts.json });
    }
    /**
     * Output that current conditions do not permit a preview release.
     */
    function invalidBranchForPreRelease() {
        Output.output(Output.createException('invalid_branch_for_pre_release', {
            summary: 'Preview releases are only supported on trunk (master) branch or branches with _open_ pull-requests. If you want to make a preview release for this branch then open a pull-request for it.',
        }), { json: opts.json });
    }
    function commitAlreadyPreAndOrStableReleased(c) {
        var _a, _b;
        const baseMessage = `You cannot make a preview release for this commit because ${c.releases.preview && c.releases.stable
            ? 'stable and preview releases were already made'
            : c.releases.preview
                ? 'a preview release was already made.'
                : 'a stable release was already made.'}`;
        if (opts.json) {
            Output.outputException('invalid_pre_release_case', baseMessage, {
                json: true,
                context: {
                    sha: c.sha,
                    preReleaseTag: (_a = c.releases.preview) === null || _a === void 0 ? void 0 : _a.version,
                    stableReleaseTag: (_b = c.releases.stable) === null || _b === void 0 ? void 0 : _b.version,
                    otherTags: c.nonReleaseTags,
                },
            });
        }
        else {
            let message = '';
            message += baseMessage + '\n';
            message += '\n';
            message += utils_1.indentBlock4(common_tags_1.stripIndents `
        The commit is:           ${c.sha}
        ${renderTagsPresent(c)}
      `);
            Output.outputRaw(message);
        }
    }
    function dryRun(info) {
        Output.outputOk('dry_run', info);
    }
    return {
        releaseType,
        noReleaseToMake,
        invalidBranchForPreRelease,
        commitAlreadyPreAndOrStableReleased,
        dryRun,
    };
}
/**
 * Given groups of parsed tags, create a nice summary of the commit for the user
 * to read in their terminal.
 */
function renderTagsPresent(c) {
    var _a, _b, _c, _d;
    const NA = 'N/A';
    let message = '';
    message += `The stable release is:   ${_b = (_a = c.releases.stable) === null || _a === void 0 ? void 0 : _a.version, (_b !== null && _b !== void 0 ? _b : NA)}\n`;
    message += `The preview release is:  ${_d = (_c = c.releases.preview) === null || _c === void 0 ? void 0 : _c.version, (_d !== null && _d !== void 0 ? _d : NA)}\n`;
    message += `Other tags present:      ${c.nonReleaseTags.length === 0 ? NA : c.nonReleaseTags.join(', ')}\n`;
    // TODO make part of diagnostics
    // if (
    //   unexpectedOtherPreReleases.length > 0 ||
    //   unexpectedOtherStableReleases.length > 0
    // ) {
    //   message += '\nWARNING\n\n'
    //   if (unexpectedOtherStableReleases) {
    //     message +=
    //       '- Multiple stable releases appear to have been on this commit when there should only ever been 0 or 1\n'
    //   }
    //   if (unexpectedOtherPreReleases) {
    //     message +=
    //       '- Multiple preview releases appear to have been on this commit when there should only ever been 0 or 1\n'
    //   }
    //   message += '\n'
    //   message += 'This may have happened because:\n'
    //   message += '- A human manually fiddled with the git tags\n'
    //   message += '- Another tool than dripip acted on the git tags\n'
    //   message += '- There is a bug in dripip\n'
    //   message += '\n'
    //   message +=
    //     'If you think there is a bug in dripip please open an issue: https://github.com/prisma-labs/dripip/issues/new.\n'
    //   message +=
    //     'Otherwise consider manually cleaning up this commit to fix the above violated invariant(s).\n'
    // }
    return message;
}
