import * as Semver from './semver';
import * as Git from './git';
/**
 * Get the previous stable and commits since then. If there is no previous
 * stable then commits are counted from start of history.
 */
export declare function getCurrentSeries(git: Git.Simple): Promise<Series>;
export declare type Commit = {
    message: string;
    sha: string;
    nonReleaseTags: string[];
    releases: {
        stable: null | Semver.StableVer;
        preview: null | Semver.PreviewVer;
    };
};
export declare type StableCommit = {
    message: string;
    sha: string;
    nonReleaseTags: string[];
    releases: {
        stable: Semver.StableVer;
        preview: null;
    };
};
export declare type PreviewCommit = {
    message: string;
    sha: string;
    nonReleaseTags: string[];
    releases: {
        stable: null;
        preview: Semver.PreviewVer;
    };
};
export declare type MaybePreviewCommit = {
    message: string;
    sha: string;
    nonReleaseTags: string[];
    releases: {
        stable: null;
        preview: null | Semver.PreviewVer;
    };
};
export declare type UnreleasedCommit = {
    message: string;
    sha: string;
    nonReleaseTags: string[];
    releases: {
        stable: null;
        preview: null;
    };
};
export declare type Series = {
    previousStable: null | StableCommit;
    commitsSinceStable: MaybePreviewCommit[];
    previousPreview: null | PreviewCommit;
    commitsSincePreview: UnreleasedCommit[];
    current: Commit;
};
export declare type SeriesLog = [null | Git.LogEntry, Git.LogEntry[]];
/**
 * Build structured series data from a raw series log.
 */
export declare function buildSeries([previousStable, commitsSince]: SeriesLog): Series;
/**
 * Get the releases at the given commit.
 */
//# sourceMappingURL=release.d.ts.map