"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const Semver = __importStar(require("./semver"));
const Git = __importStar(require("./git"));
/**
 * Get the previous stable and commits since then. If there is no previous
 * stable then commits are counted from start of history.
 */
function getCurrentSeries(git) {
    return __awaiter(this, void 0, void 0, function* () {
        return getLog().then(buildSeries);
    });
}
exports.getCurrentSeries = getCurrentSeries;
/**
 * Get the latest stable and all subsequent commits. If no stable then all
 * commits (edge-case, new project) since-inclusive initial commit. The returned
 * tuple contains the  stable commit partitioned from the subsequent commits.
 * The stable commit may be null.
 */
function getLog() {
    var e_1, _a;
    return __awaiter(this, void 0, void 0, function* () {
        const commits = [];
        let previousStableCommit = null;
        try {
            for (var _b = __asyncValues(Git.streamLog()), _c; _c = yield _b.next(), !_c.done;) {
                const log = _c.value;
                if (log.tags.find(isStableTag)) {
                    previousStableCommit = log;
                    break;
                }
                commits.push(log);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (_c && !_c.done && (_a = _b.return)) yield _a.call(_b);
            }
            finally { if (e_1) throw e_1.error; }
        }
        return [previousStableCommit, commits];
    });
}
/**
 * Build structured series data from a raw series log.
 */
function buildSeries([previousStable, commitsSince]) {
    var _a, _b;
    if (previousStable === null && commitsSince.length === 0) {
        throw new Error(`Cannot build release series with given data. There is no previous stable release and no commits since. This would indicate an unexpected error or working with a git repo that has zero commits. The latter should be guarded by upstream checks. Therefore this is bad. There must be a bug.`);
    }
    const commitsSinceStable = commitsSince.map(c => {
        var _a, _b;
        const { message, sha } = c;
        return {
            message,
            sha,
            nonReleaseTags: c.tags.filter(isUnknownTag),
            releases: {
                stable: Semver.parse((_a = c.tags.find(isStableTag), (_a !== null && _a !== void 0 ? _a : ''))),
                preview: Semver.parsePreview((_b = c.tags.find(isPreviewTag), (_b !== null && _b !== void 0 ? _b : ''))),
            },
        };
    });
    const prevPreviewI = commitsSinceStable.findIndex(c => c.releases.preview !== null);
    let previousPreview = null;
    let commitsSincePreview = [];
    if (prevPreviewI !== -1) {
        previousPreview = commitsSinceStable[prevPreviewI];
        commitsSincePreview = commitsSinceStable.slice(0, prevPreviewI);
    }
    const previousStable2 = previousStable === null
        ? null
        : {
            sha: previousStable.sha,
            message: previousStable.message,
            nonReleaseTags: previousStable.tags.filter(isUnknownTag),
            releases: {
                stable: Semver.parse(previousStable.tags.find(isStableTag)),
                preview: Semver.parsePreview((_a = previousStable.tags.find(isPreviewTag), (_a !== null && _a !== void 0 ? _a : ''))),
            },
        };
    return {
        previousStable: previousStable2,
        commitsSinceStable,
        previousPreview,
        commitsSincePreview,
        // If there are no commits since stable and no stable that means we're on a
        // repo with no commit. This edge-case is ignored. It is assumed that it
        // will be validated for before calling this function.
        current: (_b = commitsSinceStable[0], (_b !== null && _b !== void 0 ? _b : previousStable2)),
    };
}
exports.buildSeries = buildSeries;
function isUnknownTag(tag) {
    return Semver.parse(tag) === null;
}
function isStableTag(tag) {
    const v = Semver.parse(tag);
    if (v === null)
        return false;
    return Semver.isStable(v);
}
function isPreviewTag(tag) {
    const v = Semver.parse(tag);
    if (v === null)
        return false;
    return Semver.isPreview(v);
}
//
// Private helpers
//
function findLatestStable(git) {
    return __awaiter(this, void 0, void 0, function* () {
        return Git.findTag(git, { matcher: isStableTag });
    });
}
// const zeroReleases = {
//   stable: null,
//   preview: null,
// }
/**
 * Get the releases at the given commit.
 */
// export async function getReleasesAtCommit(
//   sha: string
// ): Promise<CommitReleases> {
//   const git = createGit()
//   const tags = await Git.gitGetTags(git, { ref: sha })
//   if (isEmpty(tags)) return zeroReleases
//   const parsedTags = groupByProp(tags.map(parseTag), 'type')
//   const stableTags = parsedTags.stable_release ?? []
//   const previewtags = parsedTags.pre_release ?? []
//   const invariantViolations = []
//   if (stableTags.length > 1) {
//     invariantViolations.push(
//       `Multiple stable releases found but there should only be 0 or 1: ${stableTags}`
//     )
//   }
//   if (previewtags.length > 1) {
//     invariantViolations.push(
//       `Multiple preview releases found but there should only be 0 or 1: ${stableTags}`
//     )
//   }
//   if (invariantViolations.length > 0) {
//     throw new Error(
//       `While getting the pre-existing releases at commit ${sha} the following invariant violations were found:\n\n${invariantViolations
//         .map(x => `    - ${x}`)
//         .join('\n')}`
//     )
//   }
//   return {
//     stable: stableTags[0]
//       ? {
//           type: 'stable',
//           version: stableTags[0].value,
//           sha,
//         }
//       : null,
//     preview: previewtags[0]
//       ? {
//           type: 'preview',
//           version: previewtags[0].value,
//           sha,
//           buildNum: previewtags[0].value.prerelease[1] as number,
//         }
//       : null,
//   }
// }
// export type ParsedTag =
//   | { type: 'unknown'; value: string }
//   | { type: 'stable_release'; value: SemVer.SemVer }
//   | { type: 'pre_release'; value: SemVer.SemVer }
// export function parseTag(rawTag: string): ParsedTag {
//   const semverParseResult = SemVer.parse(rawTag)
//   if (semverParseResult !== null) {
//     if (semverParseResult.prerelease.length > 0) {
//       return { type: 'pre_release', value: semverParseResult } as const
//     } else {
//       return { type: 'stable_release', value: semverParseResult } as const
//     }
//   } else {
//     return { type: 'unknown', value: rawTag } as const
//   }
// }
